(
	/*
		Notes/Maybe To-do for future versions:
			- Check/escape files names with invalid characters
			- Improve pre-export LOD crash issues checks
			- Manual add/remove filesets
			- Manual add/remove objects from filesets
	*/

	-- Clean the maxscript listener for testing purpose
	clearListener()

	-- Declare important variables
	local rolloutExport
	local sPathINI = (getFileNamePath (getThisScriptFilename()))+"\\akntoolkit.ini"
	local aFilesets = #() -- Main container of filesets and nodes
	local aUIObjectList = #() -- Index container of current select filesets objects
	local aTempObjects = #() -- Temporary container of nodes for use in functions
	local aTempList = #() -- Temporary container of nodename for use in functions

	local iRolloutPosX = getINISetting sPathINI "Settings_Max_Export_FBX" "rollout_pos_x"
	local iRolloutPosY = getINISetting sPathINI "Settings_Max_Export_FBX" "rollout_pos_y"
	-- Get the dialog windows position from ini file
	iRolloutPosX = if iRolloutPosX == "" then 100 else (iRolloutPosX as Integer)
	iRolloutPosY = if iRolloutPosY == "" then 100 else (iRolloutPosY as Integer)


	-- ===============================================================================================================
	-- == START - General shared functions
	-- ===============================================================================================================

	-- returns a string which consist of str repeated x amount times
	fn repeatStr str amount =
	(
		local s = ""
		if amount > 0 do (
			str = str as string
			for x = 1 to amount do
				s += str
		)
		return s
	)
	
	-- Display array in listener in indent form, mainly for testing purpose to view arrays
	-- formatArray [array] [level indent, usually start with 0]
	fn formatArray aArray iLevel =
	(
		if classOf aArray == Array then
		(
			format ("%#(\n") (repeatStr "    "  iLevel)
			for i in aArray do
			(
				formatArray i (iLevel+1)
			)
			format ("%)#\n") (repeatStr "    "  iLevel)
		) else (
			format ((repeatStr "   " (iLevel+1))+"%\n") aArray
		)
	)

	-- getRoot node : Finds the top most parent of the submitted node, snippet by denisT
	fn getRoot node = if isvalidnode node do 
	(
		while node.parent != undefined do (
			node = node.parent
		)
		node
	)

	-- ===============================================================================================================
	-- == END - General shared functions
	-- ===============================================================================================================

	-- Set the nodes parent groups as open/closed and returned the affected groups
	-- state : true = set group open, false = set group closed
	fn setParentGroupOpen node bState = if isvalidnode node do 
	(
		-- Is this nodecollection:#() needed?
		local aGroups = #()
		while node.parent != undefined do (
			if (isGroupHead node.parent == true and isOpenGroupHead node.parent != bState) do
			(
				setGroupOpen node.parent bState
				append aGroups (node.parent)
			)
			node = node.parent
		)
		aGroups
	)

	-- Return a nodes collection from supplied nodes, and returns unique parents (no node which is a child of another node in the collection)
	fn filterUniqueParentNodes aNodes = 
	(
		if aNodes.count > 0 then (
			local aNamelist = #()
			local n = #()
			local bCheck = false
			-- Ensure that all submitted nodes are valid nodes
			aNodes = for items in aNodes where isvalidnode items collect items
			-- Create a name list for comparison
			aNamelist = for items in aNodes collect items.name
			-- Loop through all the nodes and check them
			for i in aNamelist.count to 1 by -1 do (
				n = aNodes[i]
				bCheck = false
				-- Loop through the nodes parents and check them against the name list
				while n.parent != undefined and bCheck == false do (
					if ((findItem aNamelist n.parent.name) != 0) do (
						bCheck = true
					)
					n = n.parent
				)
				-- If a parent node was found in list, delete the current node from it
				if (bCheck) do
					aNodes = deleteItem aNodes i
			)
		)
		aNodes
	)

	-- Go through collection of parent nodes and if they are not a group, check if any of the children has the LOD_controller assigned to them, and return the found nodes
	fn validateLODs aNodes =
	(
		local nInvalidLODNodes = #() -- Collection of nodes with LOD_Controllers on them and parent isn't a group
		for n in aNodes do
		(
			--If the parent is a group and doesn't have lod on it, or isn't a group, check if the children has LOD on them, if so, throw up a flag
			if (isGroupHead n[1] == false) do (
				join nInvalidLODNodes (for c in n[2] where classOf(getVisController c[1]) == LOD_Controller collect c[1])
				-- Check if the parent node has LOD on it and isn't a group
				if (classOf(getVisController n[1]) == LOD_Controller) do (
					appendifunique nInvalidLODNodes n[1]
				)
			)
		)
		nInvalidLODNodes -- Return the found nodes
	)
	
	-- Check if the nodename is a valid collision name: [UCX/UBX/UCP/USP]_objectname[optional: _%%%]
	fn nodeIsCollisionBox sNodename =
	(
		local f = dotnetobject "System.Text.RegularExpressions.Regex" ("(U(((C|B)X)|((C|S)P))_*)")
		f.isMatch (sNodename)
	)

	-- Using the submitted node to find related collision models
	fn getRelatedCollisionBoxes node =
	(
		-- Get all collisions that follow the naming  CollisionType_ObjectName  or  CollisionType_ObjectName_Number
		local f = dotnetobject "System.Text.RegularExpressions.Regex" ("(U(((C|B)X)|((C|S)P))_" + node.name +"(_\\d+)?$)")
		for n in objects where f.isMatch (n.name) collect n
	)

	-- Get the children/childrens children of the supplied nodes and indent level, and return them as an array: #(node, indent, isChildOfParent)
	fn getChildren aNodes iLevel bIsChildOfParent =
	(
		local bIsCollisionBox
		for n in aNodes do
		(
			bIsCollisionBox = nodeIsCollisionBox(n.name)
			if (bIsCollisionBox and rolloutExport.chk_include_collision.checked) or (bIsCollisionBox == false) do (
				append aTempObjects #(n, iLevel, bIsChildOfParent)
				append aTempList (n.name)
				
				if (rolloutExport.rdo_getFromSelection.state != 3) and n.children.count > 0 do
					getChildren (n.children) (iLevel+1) true
			)

		)
	)

	-- Generate the filesets and all related objects
	fn getFilesets selectedObjs =
	(
		aFilesets = #() -- Empty the fileset container
		local aParent_Nodes = #() -- collection of parent nodes

		-- Depending on radiobutton selection, get unique parent objects for selected
		case rolloutExport.rdo_getFromSelection.state of (
			1: (
				-- get unique parent roots from selection
				for n in selectedObjs where isValidNode n do (
					appendifunique aParent_Nodes (getRoot n)
				)
			)
			2: (
				-- Filter out selections that are children of other nodes in selection, return parents only
				aParent_Nodes = (filterUniqueParentNodes selectedObjs)
			)
			3: (
				-- Make all selections their own filesets
				for n in selectedObjs where isValidNode n do
					appendifunique aParent_Nodes n
			)
		)
		-- Get children and collisions (if needed) for each parent
		for n in aParent_Nodes do (
			aTempObjects = #() -- (#(node, level, isChildOfPArent))
			aTempList = #()
			getChildren #(n) 0 false
			
			-- Now find all the collisions which belong to
			if (rolloutExport.chk_include_collision.checked) do
			(
				local outside_collisions =#() -- Found collisions which isn't child under parent
				for obj in aTempObjects do
				(
					for i in (getRelatedCollisionBoxes obj[1]) where findItem aTempList i.name == 0 do
					(
						append outside_collisions #(i, 0, false)
						-- add name to templist
						append aTempList i.name
					)
				)
				-- merge found collisions into the object list
				join aTempObjects outside_collisions
			)
			append aFilesets #(n, aTempObjects)
		)
		-- empty the temporary arrays before returning the fileset array1	

		aTempObjects = #()
		aTempList = #()
		aFilesets
	)

	-- Populate the fileset UI
	fn populate_UI_FilesetList aNodes = (
		-- Empty the fileset list ui
		rolloutExport.mlb_filesets.items = #()
		-- Populate the fileset list ui with provided nodes collection
		for i in aNodes do
			rolloutExport.mlb_filesets.items = append rolloutExport.mlb_filesets.items (i[1].name + ".fbx")
	)

	-- Populate the object list UI with the objects from the selected filesets
	-- aSelectedFilesets = selected filesets from the ui
	fn populate_UI_ObjectList aSelectedFilesets =
	(
		rolloutExport.mlb_object_list.items = #() -- Empty the object list ui
		-- Check that selected filesets isn't empty
		if (aSelectedFilesets.count > 0 ) do (		
			local sIndentFiller = "    "
			local sIndentLevel
			local bFirstSet = true
			local sListname

			aUIObjectList = #() -- Container for current objects in the object list ui

			for s in aSelectedFilesets do
			(
				-- if there is more then one fileset selected, inset fileset divider for clarity
				if aSelectedFilesets.count > 1 do (
					append aUIObjectList #(s, 1)
					rolloutExport.mlb_object_list.items = append rolloutExport.mlb_object_list.items \
						("=====>  " + aFilesets[s][1].name + ".fbx  <=====")
				)
				local x = 1
				for n in aFilesets[s][2] do
				(
					append aUIObjectList #(s, x)
					-- Check if node still exist in scene, if so, create entry with name
					if isValidNode n[1] then
					(
						sListname = (n[1].name as string)
						if (isGroupHead n[1] and (rolloutExport.chk_show_type.checked)) then
							sListname = "[" + sListname + "]"
						else if (isGroupHead n[1] == false and (rolloutExport.chk_show_type.checked)) do
							sListname += "   {"+((classOf n[1]) as string)+"}"
					) else (
						-- Node is missing, either deleted or renamed, show as deleted
						sListname = "*** deleted/missing object ****"
					)

					if (rolloutExport.chk_show_hierarchy.checked) then
					(
						sListname = (repeatStr sIndentFiller n[2])+sListname
					)
					rolloutExport.mlb_object_list.items = append rolloutExport.mlb_object_list.items sListname
					x += 1
				)
				bFirstSet = false
			)
		)
	)

	-- Set the path for where the exported files should be saved
	fn pick_export_path sNewPath =	
	(
		if (sNewPath == undefined) do (
			local sNewPath = getSavePath caption:"Select export path" initialDir:rolloutExport.edt_export_path.text
		)
		if (sNewPath != undefined and sNewPath != "") do (
			if (doesFileExist sNewPath) and (getfileattribute sNewPath #directory) then 
			(
				rolloutExport.edt_export_path.text = sNewPath
				setINISetting sPathINI "Settings_Max_Export_FBX" "export_path" sNewPath
			) else (
				messageBox "The selected path doesn't exist." title:"export path error"
				pick_export_path undefined
			)
		)
	)
	
	-- Initialize the UI to the previous session settings
	fn initiate_ui =
	(
		-- Get the last path used
		local sVal = getINISetting sPathINI "Settings_Max_Export_FBX" "export_path"

		if (sVal == "") do (
			sVal = maxfilepath
			setINISetting sPathINI "Settings_Max_Export_FBX" "export_path" (sVal as string)
		)
		rolloutExport.edt_export_path.text = sVal;
		
		sVal = getINISetting sPathINI "Settings_Max_Export_FBX" "get_from_selection"
		rolloutExport.rdo_getFromSelection.state = (if sVal == "" then 1 else sVal as integer)

		rolloutExport.chk_include_collision.checked = if (getINISetting sPathINI "Settings_Max_Export_FBX" "include_collision") == "0" then false else true
		rolloutExport.chk_show_hierarchy.checked = if (getINISetting sPathINI "Settings_Max_Export_FBX" "show_hierarchy") == "0" then false else true
		rolloutExport.chk_show_type.checked = if (getINISetting sPathINI "Settings_Max_Export_FBX" "show_type") == "0" then false else true
		
		sVal = getINISetting sPathINI "Settings_Max_Export_FBX" "export_which_filsets"
		rolloutExport.rdo_export_which_filsets.state = (if sVal == "" then 1 else sVal as integer)

		rolloutExport.chk_warning_lods.checked = if (getINISetting sPathINI "Settings_Max_Export_FBX" "warning_lods") == "0" then false else true
		rolloutExport.chk_moveto000.checked = if (getINISetting sPathINI "Settings_Max_Export_FBX" "modeto000") == "0" then false else true
	)

	fn export_FileSets aObjSets =
	(
		if (queryBox "Do you want to export selected filesets?" title:"Export Selected") do
		(
			local showExportWindow = true
			local bFirstSet = true
			local current_selected_objects = getCurrentSelection() -- Get current selection so it can be reselected after export

			max select none -- deselect all objects in scene

			-- Loop though all filesets and export them to fbx
			for s in aObjSets do (
				-- get all objects that is in the fileset and needs to be exported
				local aNodes = #()
				local aClosedGroups = #() -- 
				for i in s[2] do
				(
					append aNodes i[1]
					
					-- Find all the parent groups the node is part of and open them, and collect them to be reverted later
					if isGroupMember i[1] and isOpenGroupMember i[1] == false do
					(
						join aClosedGroups (setParentGroupOpen i[1] true)
					)
				)

				-- The filename and path for where the fbx should be saved
				local sExportFilename = rolloutExport.edt_export_path.text + @"\"+s[1].name+@".fbx"

				-- If checkbox checked, move all set objects to [0,0,0] before export
				if (rolloutExport.chk_moveto000.checked) do
				(
					local relocate_objects = for i in s[2] where i[3] == false collect #(i[1], i[1].pos, (i[1].pos - s[1].pos))
					for n in relocate_objects do
						n[1].pos = n[3]
				)

				if aNodes.count > 0 do
				(
					select aNodes
					-- Export the selected objects into a fbx files
					if bFirstSet == true and showExportWindow == true then
						exportFile sExportFilename selectedOnly:true
					else
						exportFile sExportFilename #noPrompt selectedOnly:true
				)
				
				bFirstSet = false

				-- If checkbox checked, move all set objects back to their previous position
				if (rolloutExport.chk_moveto000.checked) do
					for n in relocate_objects do
						n[1].pos = n[2]
				-- Close the affected groups again
				for i in aClosedGroups do setGroupOpen i false
			)
			-- Select the current selections again
			select current_selected_objects
		)
	)
	
	rollout rolloutExport "Mass Export FBX to UE4"
	(
		label 'lbl_exportpath' "" pos:[10,7] width:68 height:15 align:#left
		edittext 'edt_export_path' "" pos:[6,5] width:438 height:21 align:#left
		button 'btn_pick_export_path' "..." pos:[450,5] width:21 height:22 align:#left

		GroupBox 'grp_collect_from_selection' "Collect which objects from selection" pos:[5,31] width:470 height:66 align:#left
			radiobuttons 'rdo_getFromSelection' "" pos:[11,46] width:240 height:32 labels:#("Root and all children", "Children Only", "One object per file") columns:2 align:#left

			checkbox 'chk_show_type' "Show object type?" pos:[357,353] width:114 height:19 align:#left
			button 'btn_get_filesets' "Get fileset" pos:[381,45] width:86 height:22 toolTip:"Create new filesets from current selection in scene" align:#left
			button 'btn_refresh_filesets' "Refresh fileset" pos:[381,70] width:86 height:22 toolTip:"Refreshes the filesets in case of name changes (doesn't update from selection)" align:#left
		
		GroupBox 'grp_filesets_objects' "" pos:[5,94] width:470 height:280 align:#left
			multiListBox 'mlb_filesets' "Filesets" pos:[10,104] width:161 height:17 align:#left	
			multiListBox 'mlb_object_list' "Objects in selected fileset(s)" pos:[177,104] width:292 height:17 align:#left
			checkbox 'chk_include_collision' "Include related collision meshes?" pos:[11,77] width:181 height:19 align:#left
			checkbox 'chk_show_hierarchy' "Show hierarchy?" pos:[239,353] width:100 height:19 align:#left

		GroupBox 'grp_export_settings' "Export settings" pos:[6,375] width:470 height:67 align:#left
			radiobuttons 'rdo_export_which_filsets' "" pos:[15,390] width:251 height:16 labels:#("All filesets", "Selected fileset") columns:2 align:#left
			checkbox 'chk_moveto000' "Move parents to [0,0,0] on export and back?" pos:[15,404] width:239 height:17 checked:true align:#left		
			checkbox 'chk_warning_lods' "Check for redundant LOD attribute before exporting? (not 100% reliable)" pos:[15,421] width:390 height:19 align:#left		
			button 'btn_export' "Export" pos:[380,392] width:90 height:22 align:#left
		
		on rolloutExport close do
		(
			-- Save dialog current position to ini
			local aDialogPosition = (GetDialogPos rolloutExport)
			setINISetting sPathINI "Settings_Max_Export_FBX" "rollout_pos_x" ((aDialogPosition[1] as Integer) as string)
			setINISetting sPathINI "Settings_Max_Export_FBX" "rollout_pos_y" ((aDialogPosition[2] as Integer) as string)
		)

		on edt_export_path entered sNewPath do
		(
			pick_export_path sNewPath
		)

		on btn_pick_export_path pressed do
		(
			pick_export_path undefined
		)


		on rdo_getFromSelection changed bState do
		(
			setINISetting sPathINI "Settings_Max_Export_FBX" "get_from_selection" (bState as string)
		)

		on chk_include_collision changed bState do
		(
			setINISetting sPathINI "Settings_Max_Export_FBX" "include_collision" (if bState then "1" else "0")

		)

		on chk_show_hierarchy changed bState do
		(
			setINISetting sPathINI "Settings_Max_Export_FBX" "show_hierarchy" (if bState then "1" else "0")
			-- Populate the fileset object list with current selection in filesets (if any exist)
			populate_UI_ObjectList(mlb_filesets.selection as array)
		)

		on chk_show_type changed bState do
		(
			setINISetting sPathINI "Settings_Max_Export_FBX" "show_type" (if bState then "1" else "0")

			populate_UI_ObjectList(mlb_filesets.selection as array)
		)

		on btn_get_filesets pressed do
		(
			-- If confirmed, update the fileset list with the new selection from the scene
			if (queryBox "Do you want to generate new filesets from current selection?" title:"Confirm replace current filesets") do (
				local aNodes = getFilesets (getCurrentSelection())
				-- Empty the fileset and objectset ui list
				populate_UI_FilesetList(aNodes)
				populate_UI_ObjectList(#())
			)
		)

		on btn_refresh_filesets pressed do
		(
			-- Refresh the nodes in the current fileset list (in case names has changed)
			local n = for i in aFilesets collect i[1]
			local aNodes = getFilesets(n)
			-- Empty the fileset and objectset ui list
			populate_UI_FilesetList(aNodes)
			populate_UI_ObjectList(#())
		)
		
		on mlb_filesets selected bState do (
			-- Update the object list with the selected filesets
			local s = mlb_filesets.selection as array
			populate_UI_ObjectList(s)
		)

		on mlb_object_list rightClick do (
			-- When rightclicking the object list box, it will select those objects in scene
			-- Cuurently this only works with one fileset selected, unfortunately. Might add later.
			local s = (mlb_object_list.selection as array)
			local nSelect = #()
			if (s.count > 0) do (
				for n in s do
				(
					fs = aUIObjectList[n][1]
					os = aUIObjectList[n][2]
					nObj = aFilesets[fs][2][os][1]
					if isValidNode nObj do
						append nSelect nObj
				)
				if (nSelect.count > 0) do
					select nSelect
			)
		)
		
		on rdo_export_which_filsets changed bState do
		(
			setINISetting sPathINI "Settings_Max_Export_FBX" "export_which_filsets" (bState as string)
		)
		
		on chk_warning_lods changed bState do
		(
			setINISetting sPathINI "Settings_Max_Export_FBX" "warning_lods" (if bState then "1" else "0")
		)

		on chk_moveto000 changed bState do
		(
			setINISetting sPathINI "Settings_Max_Export_FBX" "moveto000" (if bState then "1" else "0")
		)

		on btn_export pressed do
		(	
			local aExportSets = #()
			if rdo_export_which_filsets.state == 1 then (
				-- get all filesets
				aExportSets = aFilesets
			) else if mlb_filesets.selection.count > 0 do (
				-- get only selected filesets
				for i in mlb_filesets.selection do
					append aExportSets (aFilesets[i])
			)
		
			-- If there is any filesets selected for export, continue
			if (aExportSets.count > 0) do
			(
				-- If checked, validate all objects with LOD attributes are part of a group
				local aLODcheck = #()
				if (chk_warning_lods.checked) do (
					local aLODcheck = (validateLODs aExportSets) as array
				)
				
				-- If any conflicting lods was found, prompt a warning, else start exporting
				if (aLODcheck.count > 0) then (
					local q = (yesNoCancelBox  (aLODcheck.count as string + " object(s) has LOD attribute however parent isn't a group.\nDo you want to select the objects with the redundant LOD attribute?\n\n[YES], select the objects\n[NO], continue exporting (might result in crash)\n[CANCEL] the export\n") title:"LOD issue found")
					if (q == #yes) then (
						-- select the objects which have the redundent LOD attribute assigned to them
						select aLODcheck
					) else if (q == #no) do (
						-- Export the filesets
						export_FileSets aExportSets
					)
				) else (
					-- Export the filesets
					export_FileSets aExportSets
				)
			)
		)
	)

	-- Create the dialog and initate the ui values
	try(destroyDialog rolloutExport)catch()	
	createdialog rolloutExport 480 446 iRolloutPosX iRolloutPosY lockHeight:true
	initiate_ui()

	-- Automatically populate the filesets with current selection on opening window
	if ((getCurrentSelection()).count > 0) do
	(
		local aNodes = getFilesets (getCurrentSelection())
		-- Empty the fileset and objectset ui list
		populate_UI_FilesetList(aNodes)
		populate_UI_ObjectList(#())
	) 
)